from typing import Dict

from woodchipper.context import LoggableValue


class BaseMonitor:
    def __init__(self):
        raise NotImplementedError()

    def setup(self):
        raise NotImplementedError()

    def finish(self) -> Dict[str, LoggableValue]:
        raise NotImplementedError()



import contextvars
import inspect
import os
import time
from collections.abc import MutableMapping
from decimal import Decimal
from functools import wraps
from typing import Any, Mapping, NamedTuple, Optional, Tuple, Union, cast

import woodchipper

LoggableValue = Optional[Union[str, int, bool, Decimal, float]]
LoggingContextType = Mapping[str, LoggableValue]


class Missing:
    """A sentinel value to be raised if something is missing."""

    def __repr__(self):
        return "<missing>"


# Poor mans singleton
missing = Missing()


def _split_head_node(str_path: str, delimiter=".") -> Tuple[str, str]:
    head, tail = str_path.split(delimiter, maxsplit=1) if delimiter in str_path else (str_path, "")
    return head, tail


def _get_key_or_attr(obj, name: str, missing=missing):
    try:
        return getattr(obj, name)
    except AttributeError:
        try:
            return obj[name]
        except (TypeError, KeyError):
            return missing


def pluck_value(obj, str_path: str, delimiter=".") -> Union[Any, Missing]:
    """Dig into an object or dict Using a JMSE-like path syntax. Return Missing if value cannot be found. Only
    supports mapping-like behavior so cannot search sequences."""
    if not str_path:
        return obj
    else:
        head, remaining = _split_head_node(str_path, delimiter)
        value = _get_key_or_attr(obj, head)
        if value is missing:
            return missing
        else:
            return pluck_value(value, remaining, delimiter)


class ParamSearchConfig(NamedTuple):
    dig_path: str
    logvar_name: str
    diggable: bool


def _build_path_head_to_param_config_map(kwargs: Mapping, *, delimiter: str = ".") -> Mapping[str, ParamSearchConfig]:
    return_mapping = {}
    for logvar_name, dig_path in kwargs.items():
        if dig_path == "":
            raise ValueError("Dig path cannot be an empty string")

        dig_path_head, dig_path_tail = _split_head_node(dig_path, delimiter)

        return_mapping[dig_path_head] = ParamSearchConfig(
            dig_path=dig_path_tail,
            logvar_name=logvar_name,
            diggable=True if dig_path_tail else False,
        )

    return return_mapping


class LoggingContextVar(MutableMapping):
    _var: contextvars.ContextVar[LoggingContextType]

    def __init__(self, name):
        self._var = contextvars.ContextVar(name, default={})

    def __getitem__(self, key: str) -> LoggableValue:
        return self._var.get().get(key)

    def __setitem__(self, key: str, value: LoggableValue) -> contextvars.Token:
        updated = dict(self._var.get())
        updated[key] = value
        return self._var.set(updated)

    def __eq__(self, val) -> bool:
        return val == self._var.get()

    def __delitem__(self, key: str) -> contextvars.Token:
        updated = dict(self._var.get())
        del updated[key]
        return self._var.set(updated)

    def __iter__(self):
        return self._var.get().__iter__()

    def __len__(self):
        return self._var.get().__len__()

    def __repr__(self):
        return repr(self._var.get())

    def as_dict(self):
        return dict(self._var.get())

    def update(self, d: LoggingContextType) -> contextvars.Token:
        token = None
        for key, value in d.items():
            new_token = self.__setitem__(key, value)
            if token is None:
                token = new_token
        if token is None:
            raise ValueError("Cannot update with empty mapping.")
        return token

    def reset(self, token: contextvars.Token):
        self._var.reset(token)


def _convert_to_loggable_value(value: Any) -> LoggableValue:
    loggable_types = {str, int, bool, Decimal, float, None}
    if type(value) in loggable_types:
        return cast(LoggableValue, value)
    else:
        try:
            return str(value)
        except Exception:
            raise ValueError(
                f"Value type (type={type(value)} is not a not a LoggableType ({str, int, bool, Decimal, float, None}) "
                f"and cannot be converted to string."
            )


class LoggingContext:
    """A context manager for logging context. Can also be used as a decorator.

    Usage:
    ```python
    with LoggingContext(data_to_inject_to_logging_context):
        some_function()
    ```
    """

    start_time: float

    def __init__(
        self, name=None, *, _prefix=missing, _missing_default=missing, _path_delimiter=".", **kwargs: LoggableValue
    ):
        self.name = name
        self.injected_context = kwargs
        self.prefix = os.getenv("WOODCHIPPER_KEY_PREFIX") if _prefix is missing else _prefix
        self._token = None
        self._monitors = [cls() for cls in woodchipper._monitors]
        self.missing_default = _missing_default
        self.path_delimiter = _path_delimiter

    def __enter__(self):
        self._token = logging_ctx.update(
            {(f"{self.prefix}.{k}" if self.prefix else k): v for k, v in self.injected_context.items()}
        )
        for monitor in self._monitors:
            monitor.setup()
        self.start_time = time.time()

    def __exit__(self, exc_type, exc_val, exc_tb):
        current_frame = inspect.currentframe()
        calling_frame = inspect.getouterframes(current_frame, 2)[1]
        module = inspect.getmodule(calling_frame[0])
        module_name = module.__name__ if module else "<unknown>"
        # If self.name is None, derive the name from the module and function from which we were called
        if self.name is None:
            func_name = calling_frame[3]
            self.name = f"{module_name}:{func_name}"

        monitored_data: LoggingContextType = {}
        monitored_data.update({"context.time_to_run_musec": int((time.time() - self.start_time) * 1e6)})
        for monitor in self._monitors:
            monitored_data.update(monitor.finish())
        woodchipper.get_logger(module_name).info(
            f"Exiting context: {self.name}", context_name=self.name, **monitored_data
        )
        assert self._token
        logging_ctx.reset(self._token)
        self._token = None
        return False

    def __call__(self, f):
        self.decorator_mapping = _build_path_head_to_param_config_map(
            self.injected_context, delimiter=self.path_delimiter
        )
        if self.name is None:
            module_name = f.__module__ or "<unknown>"
            self.name = f"{module_name}:{f.__name__}"

        @wraps(f)
        def wrapper(*func_args, **func_kwargs):

            # This will pair up the function signature parameters with the arguments that were passed. It's ...amazing.
            # If we wanted to apply defaults as well (that weren't passed but were defined as defaults in the
            # signature, there's a fairly easy process for that too through this api as well).

            # Note: We not will dig into the *args or **kwargs of the function, if the function definition used
            # *args or **kwargs. It's more difficult for **kwargs and not possible with *args.
            # TODO: attempt digging into function **kwargs as well?
            # TODO: consider a configurable 'log_defaults' decorator parameter that will allow the logging of
            #  function definition defaults when those arguments aren't passed to the function
            mapped_args = inspect.signature(f).bind(*func_args, **func_kwargs).arguments

            self.injected_context = {}

            # TODO: extract this functionality so that it's unit testable
            for dec_param_name, param_config_entry in self.decorator_mapping.items():
                if dec_param_name not in mapped_args:
                    raw_value = self.missing_default
                else:
                    if param_config_entry.diggable:
                        raw_value = pluck_value(
                            mapped_args[dec_param_name],
                            param_config_entry.dig_path,
                            delimiter=self.path_delimiter,
                        )
                    else:
                        raw_value = mapped_args[dec_param_name]

                self.injected_context[param_config_entry.logvar_name] = _convert_to_loggable_value(raw_value)

            with self:
                return f(*func_args, **func_kwargs)

        return wrapper


logging_ctx = LoggingContextVar("logging_ctx")



import importlib
import logging.config
from typing import Callable, Dict, List, Set, Type, Union

import structlog

from woodchipper.monitors import BaseMonitor

_monitors: Set[Type[BaseMonitor]] = set()
_facilities: Dict[str, str] = {}


class BaseConfigClass:
    processors: List[Callable]
    factory: Callable
    wrapper_class: Type
    renderer: Callable


def configure(
    *,
    config: Union[str, Type[BaseConfigClass]] = "woodchipper.configs.DevLogToStdout",
    facilities: Dict[str, str] = {"": "INFO"},
    override_existing: bool = True,
    monitors: List[Type[BaseMonitor]] = [],
) -> None:
    _monitors.update(set(monitors))
    _facilities.update(facilities)
    if isinstance(config, str):
        module_name, cls_name = config.rsplit(".", 1)
        module_obj = importlib.import_module(module_name)
        config = getattr(module_obj, cls_name)
        assert isinstance(config, type)
    dict_config = {
        "version": 1,
        "disable_existing_loggers": override_existing,
        "formatters": {
            "structlog": {
                "()": structlog.stdlib.ProcessorFormatter,
                "foreign_pre_chain": config.processors,
                "processors": [structlog.stdlib.ProcessorFormatter.remove_processors_meta, config.renderer],
            }
        },
        "handlers": {
            "woodchipper": {
                "level": "DEBUG",
                "formatter": "structlog",
                "class": "logging.StreamHandler",
                "stream": "ext://sys.stdout",
            }
        },
        "loggers": {
            facility: {"handlers": ["woodchipper"], "level": level, "propagate": False}
            for facility, level in facilities.items()
        },
    }
    logging.config.dictConfig(dict_config)

    structlog.configure(
        processors=config.processors + [structlog.stdlib.ProcessorFormatter.wrap_for_formatter],
        wrapper_class=config.wrapper_class,
        context_class=dict,
        logger_factory=config.factory,
        cache_logger_on_first_use=False,
    )


def reset():
    structlog.reset_defaults()


def get_monitors():
    return list(_monitors)


def get_facilities():
    return _facilities


def get_logger(name: str) -> structlog.BoundLogger:
    return structlog.get_logger(name)



import os

from woodchipper.context import LoggingContext


class WoodchipperLambda:
    def __init__(self, app):
        self._app = app

    def __call__(self, environ, start_response):
        if "LAMBDA_TASK_ROOT" not in os.environ:
            # This is a decent sentinel for being in a Lambda environment
            return self._app(environ, start_response)
        context = environ.get("lambda.context", object())
        with LoggingContext(
            "awslambda:dispatch",
            **{
                "aws-request-id": getattr(context, "aws_request_id", None),
                "function-version": getattr(context, "function_version", None),
                "function-name": getattr(context, "function_name", None),
            },
            _prefix="lambda",
        ):
            return self._app(environ, start_response)



import importlib
import logging
import os

import structlog

try:
    from structlog_sentry import SentryJsonProcessor
except ImportError:
    # structlog_sentry is an optional dependency
    # if not installed, make a no-op processor
    class SentryJsonProcessor:
        def __init__(self, *args, **kwargs):
            pass

        def __call__(self, logger, name, event_dict):
            return event_dict


import woodchipper.logger
import woodchipper.processors
from woodchipper import BaseConfigClass

callsite_parameters = [
    structlog.processors.CallsiteParameter.FUNC_NAME,
    structlog.processors.CallsiteParameter.LINENO,
    structlog.processors.CallsiteParameter.MODULE,
]


# Users can override color style
if os.getenv("WOODCHIPPER_COLOR_STYLE") is not None:
    _style_mod_name, _style_var_name = os.environ["WOODCHIPPER_COLOR_STYLE"].rsplit(".", 1)
    _style_mod = importlib.import_module(_style_mod_name)
    _style_dict = getattr(_style_mod, _style_var_name)
else:
    _style_dict = None


class Minimal(BaseConfigClass):
    """
    Really used for unit tests. That's it.
    """

    processors = [
        structlog.stdlib.add_log_level,
        structlog.stdlib.add_logger_name,
    ]
    factory = structlog.stdlib.LoggerFactory()
    wrapper_class = woodchipper.logger.BoundLogger
    renderer = structlog.processors.JSONRenderer()


class DevLogToStdout(BaseConfigClass):
    processors = [
        structlog.stdlib.add_log_level,
        structlog.stdlib.add_logger_name,
        structlog.stdlib.ExtraAdder(),
        structlog.stdlib.PositionalArgumentsFormatter(),
        structlog.processors.StackInfoRenderer(),
        structlog.dev.set_exc_info,
        structlog.processors.UnicodeDecoder(),
        structlog.processors.TimeStamper(fmt="%Y-%m-%d %H:%M.%S", utc=False),
        structlog.processors.CallsiteParameterAdder(
            parameters=callsite_parameters, additional_ignores=["woodchipper"]
        ),
        woodchipper.processors.inject_context_processor,
    ]
    factory = structlog.stdlib.LoggerFactory()
    wrapper_class = woodchipper.logger.BoundLogger
    renderer = structlog.dev.ConsoleRenderer(
        colors=not os.getenv("WOODCHIPPER_DISABLE_COLORS"), level_styles=_style_dict
    )


class JSONLogToStdout(BaseConfigClass):
    processors = [
        structlog.stdlib.add_log_level,
        structlog.stdlib.add_logger_name,
        structlog.stdlib.ExtraAdder(),
        structlog.stdlib.PositionalArgumentsFormatter(),
        structlog.processors.StackInfoRenderer(),
        woodchipper.processors.GitVersionProcessor(),
        structlog.processors.TimeStamper(fmt="%Y-%m-%d %H:%M.%S", utc=False),
        structlog.processors.CallsiteParameterAdder(
            parameters=callsite_parameters, additional_ignores=["woodchipper"]
        ),
        SentryJsonProcessor(level=logging.ERROR, as_extra=True, active=not os.getenv("WOODCHIPPER_DISABLE_SENTRY")),
        structlog.processors.format_exc_info,
        structlog.processors.UnicodeDecoder(),
        woodchipper.processors.inject_context_processor,
    ]
    factory = structlog.stdlib.LoggerFactory()
    wrapper_class = woodchipper.logger.BoundLogger
    renderer = structlog.processors.JSONRenderer()



import uuid

from flask import g, request

from woodchipper.context import LoggingContext, logging_ctx

BLACKLISTED_HEADERS = ["authorization", "cookie"]


class WoodchipperFlask:
    def __init__(self, app, blacklisted_headers=BLACKLISTED_HEADERS, request_id_factory=None):
        self._app = app
        self._blacklisted_headers = blacklisted_headers
        self._request_id_factory = request_id_factory or (lambda: str(uuid.uuid4()))
        self.vanilla_full_dispatch_request = app.full_dispatch_request

    def wrapped_full_dispatch_request(self):
        if not getattr(g, "request_id", None):
            g.request_id = self._request_id_factory()
        with LoggingContext(
            "flask:request",
            **{
                "id": g.request_id,
                "body_size": request.content_length,
                "method": request.method,
                "path": request.base_url,
                **{
                    f"query_param.{param_key.lower()}": param_val_list[0]
                    if len(param_val_list) == 1
                    else param_val_list
                    for param_key, param_val_list in request.args.lists()
                },
                **{
                    f"header.{header.lower()}": ("******" if header.lower() in self._blacklisted_headers else value)
                    for header, value in request.headers.items()
                },
            },
            _prefix="http",
        ):

            try:
                response = self.vanilla_full_dispatch_request()
            except Exception:
                # non HTTPErrors will not be caught by flask, so if that's
                # happening we assume that it's crashing the response and
                # thus 500 level
                logging_ctx.update(
                    {
                        "http.response.status_code": 500,
                    }
                )
                raise
            else:

                logging_ctx.update(
                    {
                        "http.response.status_code": response.status_code,
                        "http.response.content_length": response.content_length,
                    }
                )
                return response

    def chipperize(self):
        self._app.full_dispatch_request = self.wrapped_full_dispatch_request



import json
import logging

from woodchipper import context


def inject_context_processor(logger: logging.Logger, method: str, event_dict: dict):
    """
    A function to update context with new log event key/value
    Returns updated context items
    """
    ctx_items = context.logging_ctx.as_dict()
    ctx_items.update(event_dict)

    return ctx_items


class GitVersionProcessor:
    """
    Loads the contents of a static file written during CI builds with the git revision info.
    """

    def __init__(self, version_json_path="version.json"):
        self.version_json_path = version_json_path

    def __call__(self, logger: logging.Logger, method: str, event_dict: dict):
        try:
            with open(self.version_json_path) as ifs:
                git_version = json.load(ifs)
                event_dict.update({f"git.{k}": v for k, v in git_version.items()})
        except (OSError, json.JSONDecodeError):
            pass
        return event_dict



import time
from typing import Callable, Optional

from sqlalchemy import engine, event

from woodchipper.monitors import BaseMonitor


class SQLAlchemyMonitor(BaseMonitor):
    statement_count: int
    last_statement_started_at: Optional[float]
    total_db_time: float
    engine: engine.Engine
    instance_setup_cb: Optional[Callable]

    def __init__(self):
        self.statement_count = 0
        self.last_statement_started_at = None
        self.total_db_time = 0.0

    def handle_before_event(self, conn, cursor, statement, parameters, context, executemany):
        self.last_statement_started_at = time.time()

    def handle_after_event(self, conn, cursor, statement, parameters, context, executemany):
        assert self.last_statement_started_at
        exec_time = time.time() - self.last_statement_started_at
        self.last_statement_started_at = None
        self.total_db_time += exec_time
        self.statement_count += 1

    def setup(self):
        if self.instance_setup_cb is not None:
            self.instance_setup_cb()
        event.listen(self.engine, "before_cursor_execute", self.handle_before_event)
        event.listen(self.engine, "after_cursor_execute", self.handle_after_event)

    def finish(self):
        event.remove(self.engine, "before_cursor_execute", self.handle_before_event)
        event.remove(self.engine, "after_cursor_execute", self.handle_after_event)
        return {"sql.statement_count": self.statement_count, "sql.total_db_time_musec": int(self.total_db_time * 1e6)}



import os
from typing import Any, Mapping, Optional, Sequence

from structlog.stdlib import BoundLogger as StdlibBoundLogger

NO_PREFIX_LIST = ["exc_info"]


class BoundLogger(StdlibBoundLogger):
    def prefix_kwargs(self, **kwargs: Any) -> Mapping[str, Any]:
        prefix = os.getenv("WOODCHIPPER_KEY_PREFIX")
        if not prefix:
            return kwargs
        prefixed_kw = {}
        for key, value in kwargs.items():
            prefixed_kw[key if "." in key or key in NO_PREFIX_LIST else f"{prefix}.{key}"] = value
        return prefixed_kw

    def prefix_keys(self, *keys: str) -> Sequence[str]:
        prefix = os.getenv("WOODCHIPPER_KEY_PREFIX")
        if not prefix:
            return keys
        return [key if "." in key or key in NO_PREFIX_LIST else f"{prefix}.{key}" for key in keys]

    def bind(self, **new_values: Any) -> "BoundLogger":
        return super().bind(**self.prefix_kwargs(**new_values))  # type: ignore

    def unbind(self, *keys: str) -> "BoundLogger":
        return super().unbind(*self.prefix_keys(*keys))  # type: ignore

    def try_unbind(self, *keys: str) -> "BoundLogger":
        return super().try_unbind(*self.prefix_keys(*keys))  # type: ignore

    def _proxy_to_logger(
        self, method_name: str, event: Optional[str] = None, *event_args: str, **event_kw: Any
    ) -> Any:
        return super()._proxy_to_logger(method_name, event, *event_args, **self.prefix_kwargs(**event_kw))
